/*
 *
 * hal_phy_interrupt_regs.h
 * Description:
 * defines the MAC interrupt registers and their content
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

/*! \file hal_phy_interrupt_regs.h                                         */
/*  \brief defines the MAC interrupt registers and their content.          */

#ifndef _HAL_PHY_INTERRUPT_REGS_H_
#define _HAL_PHY_INTERRUPT_REGS_H_


/**************************************************************************/
/*      INCLUDES                                                          */
/**************************************************************************/


/**************************************************************************/
/*      INTERFACE  Defines and Structs                                    */
/**************************************************************************/
#define PHY_KERNEL_BASE_ADDRESS                     0xD9070000
#define HAL_PHY_ADDRESS_SPACE_SIZE                  0x1000

#define PHY_RX_INT_OFFSET                           6

/* NOTE:                                                                  */
/* Unlike the MAC interrupts, the PHY interrupt causes are not directly   */
/* linked to the bits in the status-register. This is the definition of   */
/* PHY-Interrupt causes that will be used by all the SW mechanisms.       */

/*****************/
/* Rx1           */
/*****************/
/*0-5*/
/*PHY_RX1_INT_FEC_MONITOR - not in use               */
#define PHY_RX1_INT_FEC_LOST                         0
#define PHY_RX1_INT_FEC_LOCK                         1
#define PHY_RX1_INT_MPEG_LOST                        2
#define PHY_RX1_INT_MPEG_LOCK                        3
#define PHY_RX1_INT_QAM_LOCK                         4
#define PHY_RX1_INT_QAM_LOST                         5


/*****************/
/* Rx2           */
/*****************/
/*6-11*/
/*PHY_RX2_INT_FEC_MONITOR - not in use               */
#define PHY_RX2_INT_FEC_LOST                         6
#define PHY_RX2_INT_FEC_LOCK                         7
#define PHY_RX2_INT_MPEG_LOST                        8
#define PHY_RX2_INT_MPEG_LOCK                        9
#define PHY_RX2_INT_QAM_LOCK                         10
#define PHY_RX2_INT_QAM_LOST                         11

/*****************/
/* Rx3           */
/*****************/
/*12-17*/
/*PHY_RX3_INT_FEC_MONITOR - not in use               */
#define PHY_RX3_INT_FEC_LOST                         12
#define PHY_RX3_INT_FEC_LOCK                         13
#define PHY_RX3_INT_MPEG_LOST                        14
#define PHY_RX3_INT_MPEG_LOCK                        15
#define PHY_RX3_INT_QAM_LOCK                         16
#define PHY_RX3_INT_QAM_LOST                         17


/*****************/
/* Rx4           */
/*****************/
/*18-23*/
/*PHY_RX4_INT_FEC_MONITOR - not in use               */
#define PHY_RX4_INT_FEC_LOST                         18
#define PHY_RX4_INT_FEC_LOCK                         19
#define PHY_RX4_INT_MPEG_LOST                        20
#define PHY_RX4_INT_MPEG_LOCK                        21
#define PHY_RX4_INT_QAM_LOCK                         22
#define PHY_RX4_INT_QAM_LOST                         23


/*****************/
/* Tx1-Tx4       */
/*****************/
/*24-28*/
#define PHY_PLL_LOCKED_ENABLE                        24
#define PHY_TX1_AMP_ENABLE                           25
#define PHY_TX2_AMP_ENABLE                           26
#define PHY_TX3_AMP_ENABLE                           27
#define PHY_TX4_AMP_ENABLE                           28




/**************************************************************************/
/*      INTERFACE VARIABLES (prefix with EXTERN)                          */
/**************************************************************************/

/**************************************************************************/
/*      INTERFACE FUNCTIONS Prototypes:                                   */
/**************************************************************************/

#endif

