/* $USAGI: tftp.c,v 1.4 2001/02/15 10:04:08 yoshfuji Exp $ */

/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*----------------------------------------------------------------------------
// Copyright 2009, Texas Instruments Incorporated
//
// This program has been modified from its original operation by Texas Instruments
// to do the following:
//
// 1. Added binary exponential back off and retransmit mechanism.
// 2. Enabled TFTP Put.
// 3. Support ICMPv4 and ICMPv4.
// 4. Plug-in is informed about each timeout event.
// 5.
//
// THIS MODIFIED SOFTWARE AND DOCUMENTATION ARE PROVIDED
// "AS IS," AND TEXAS INSTRUMENTS MAKES NO REPRESENTATIONS
// OR WARRENTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
// TO, WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE OR
// DOCUMENTATION WILL NOT INFRINGE ANY THIRD PARTY PATENTS,
// COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS.
//
// These changes are covered as per original license.
//-----------------------------------------------------------------------------*/

/*
 * From: @(#)tftp.c	5.10 (Berkeley) 3/1/91
 */
char tftp_rcsid[] =
"$Id: tftp.c,v 1.10 2000/07/22 19:06:29 dholland Exp $";

/* Many bug fixes are from Jim Guyton <guyton@rand-unix> */

/*
 * TFTP User Program -- Protocol Machines
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <arpa/inet.h>

#include <netinet/in.h>
#include <arpa/tftp.h>
#include <linux/icmp.h>

#ifdef INET6
#include <linux/icmpv6.h>
#endif

#include <asm/types.h>
#include <linux/errqueue.h>

#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <setjmp.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>		/* srand() rand() */

#include "TI_tftp.h"

#include "tftpsubs.h"

#ifdef _USAGI
    #include "version.h"
#else
    #include "../version.h"
#endif

#define UPTIME_FILE_PATH        "/proc/uptime"
#define MAX_LINE_SIZE           64

extern  struct sockaddr_storage peeraddr; /* filled in by main */
extern  socklen_t peer_len;
extern  int     f;                      /* the opened socket */
extern  int     trace;
extern  int     verbose;
extern  int     rexmtval;
extern  int     maxtimeout;
extern sigjmp_buf toplevel;
extern  int     ignoreicmp;
void sendfile(int fd, char *name, char *modestr);
void recvfile(int fd, char *name, char *modestr);


static char ackbuf[PKTSIZE];
static int timeout = 1;
static sigjmp_buf timeoutbuf;


static int makerequest(int request, char *name,
                       struct tftphdr *tp, char *mode);
static void nak(int errnor, struct sockaddr_storage *peer);
static void tpacket(const char *s, struct tftphdr *tp, int n);
static void startclock(void);
static void stopclock(void);
static void printstats(const char *direction, unsigned long amount);
static int get_sysUpTime(unsigned int *sysUpTimeVal);

#define EX_ICMP_DEST_UNREACH         9

static struct msghdr msgh;
static struct iovec iov[1];
static union {
    struct cmsghdr cmh;
    char control[CMSG_SPACE(200)];
    }ctrl_un;

#define INIT_MSG_STR( databuf, datalen, addrfrom, flag ) \
   do{\
     msgh.msg_control = (void*)&ctrl_un; msgh.msg_controllen = sizeof(ctrl_un); \
     msgh.msg_flags = (flag);                                            \
     iov[0].iov_base = (void*)(databuf); iov[0].iov_len = (datalen);               \
     msgh.msg_iov = iov; msgh.msg_iovlen = 1;                          \
     msgh.msg_name = (void*)&(addrfrom); msgh.msg_namelen = sizeof(addrfrom);     \
   }while(0);

static inline int check_src(struct sockaddr_storage *from)
{
    switch ( ((struct sockaddr*)from)->sa_family ) {
        case AF_INET:
             if( ((struct sockaddr_in *)&peeraddr)->sin_addr.s_addr == ((struct sockaddr_in *)from)->sin_addr.s_addr )
                 return 0;
             break;
#ifdef INET6
        case AF_INET6:
            return 0;
            break;
#endif
        default:    break;
      }
      return -1;
}

static inline void change_port(struct sockaddr_storage *to, struct sockaddr_storage *from)
{
      switch ( ((struct sockaddr*)to)->sa_family ) {
        case AF_INET:
            ((struct sockaddr_in *)to)->sin_port = ((struct sockaddr_in *)from)->sin_port;
        break;
#ifdef INET6
        case AF_INET6:
            ((struct sockaddr_in6 *)to)->sin6_port = ((struct sockaddr_in6 *)from)->sin6_port;
        break;
#endif
        default:    break;
      }
      return;
}

/* rcvmsgopt
   returns 1 in case of ICMP error, or 0 otherwise */
static int rcvmsgopt( void )
{
    struct cmsghdr *cmsgh = NULL;
    int n, ret = 0;

    if( -1 == ( n = recvmsg(f, &msgh, MSG_ERRQUEUE)) ) {
         perror("tftp: rcvmsgopt");
    }
    for( cmsgh = CMSG_FIRSTHDR(&msgh); cmsgh != NULL; cmsgh = CMSG_NXTHDR(&msgh,cmsgh)) {

         DBG_PRINT(("==>>[%s]: cmsg_level=%d,cmsg_type=%d, cmsg_len=%d\n",\
                   __func__, cmsgh->cmsg_level, cmsgh->cmsg_type, cmsgh->cmsg_len ))

         if( (SOL_IP == cmsgh->cmsg_level && cmsgh->cmsg_type == IP_RECVERR)
#ifdef INET6
           ||(SOL_IPV6 == cmsgh->cmsg_level && cmsgh->cmsg_type == IPV6_RECVERR)
#endif
                                                                                 ) {

			 struct sock_extended_err  *ee = (struct sock_extended_err*)CMSG_DATA(cmsgh);

             if( ee->ee_origin == SO_EE_ORIGIN_ICMP  && ee->ee_type == ICMP_DEST_UNREACH ) {
				 if( ee->ee_code >= ICMP_NET_UNREACH &&  ee->ee_code <= ICMP_PREC_CUTOFF ) {
					 DBG_PRINT(( "==>>[%s]: ICMP error <%.2d> - return 1\n", __func__, ee->ee_code ))
                     return 1;
				 }
			 }
#ifdef INET6
             else if (ee->ee_origin == SO_EE_ORIGIN_ICMP6 && ee->ee_type == ICMPV6_DEST_UNREACH ) {
				 if( ee->ee_code >= ICMPV6_NOROUTE &&  ee->ee_code <= ICMPV6_PORT_UNREACH ) {
					 DBG_PRINT(( "==>>[%s]: ICMPv6 error <%.2d> - return 1\n", __func__, ee->ee_code ))
					 return 1;
				 }
             }
#endif	/* #ifdef INET6 */
			 DBG_PRINT(( "==>>[%s]: receive unknown ICMP error <%.2d>\n", __func__, ee->ee_code ))
		} /* if( cmsgh->cmsg_level && cmsgh->cmsg_type ) */
    }
    return ret;
}

static
void
timer(int signum)
{
    int xmitcap;

    (void)signum;

    if( rcvmsgopt() ) {
		if(!ignoreicmp){
        TIplug.tftp_report_config(REP_SET + EX_ICMP_DEST_UNREACH, "Destination unreachable");
        siglongjmp(toplevel, -1);
    }
	}
    if ( TI_error == 0 )    {
        /* RRQ acked but DATA timed out */
        TIplug.tftp_report_config(REP_SET + EX_DATAERR, "DATA time out");
    }

    xmitcap = rexmtval<<1;        /* limit retransmit to be < 16 sec */
    timeout++;
    if ( xmitcap > 16 )
        xmitcap = 16;
    rexmtval = xmitcap;
    if ( rexmtval == 0 )
        rexmtval++;
    if ( verbose )
        printf("to: %d, rex: %d, cap: %d, max: %d\n", timeout, rexmtval, xmitcap, maxtimeout);

    if ( timeout > maxtimeout ) {
        printf("Transfer timed out.\n");
        if ( TI_error == -1 ) {    /* DATA not yet started */
            /* RRQ not acked */
            TIplug.tftp_report_config(REP_SET + EX_UNAVAILABLE, "RRQ time out");
        }
        else if ( TI_error > 0 ) {
            /* Received ENOTFOUND error prob */
            TIplug.tftp_report_config(REP_SET + TI_error, "ENOTFOUND error prob");
        }
        else {
            /* RRQ acked but DATA timed out */
            TIplug.tftp_report_config(REP_SET + EX_DATAERR, "DATA time out");
        }

        if ( TIplug.tftp_timeout_cb )
            TIplug.tftp_timeout_cb();   /* plugin call for failed get */

        siglongjmp(toplevel, -1);
    }
    siglongjmp(timeoutbuf, 1);
}

/*
 * Send the requested file.
 */
void
sendfile(int fd, char *name, char *mode)
{
    register struct tftphdr *ap;       /* data and ack packets */
    struct tftphdr *dp;
    volatile int size = 0;
    volatile u_int16_t block = 0;
    int n;
    volatile unsigned long amount = 0;
    struct sockaddr_storage peer, from;
    socklen_t fromlen;
    volatile int convert;            /* true if doing nl->crlf conversion */
    FILE *file;
    volatile int firsttrip = 1;

    startclock();           /* start stat's clock */
    dp = r_init();          /* reset fillbuf/read-ahead code */
    ap = (struct tftphdr *)ackbuf;
    file = fdopen(fd, "r");
    convert = !strcmp(mode, "netascii");
    memcpy(&peer, &peeraddr, peer_len);

    mysignal(SIGALRM, timer);
    do {
	if (firsttrip) {
	    size = makerequest(WRQ, name, dp, mode) - 4;
	}
	else {
	    /*      size = read(fd, dp->th_data, SEGSIZE);   */
	    size = readit(file, &dp, convert);
	    if (size < 0) {
		nak(errno + 100, &peer);
		break;
	    }
	    dp->th_opcode = htons((u_short)DATA);
	    dp->th_block = htons((u_short)block);
	}
	timeout = 0;
	(void) sigsetjmp(timeoutbuf, 1);
      send_data:
	if (trace)
	    tpacket("sent", dp, size + 4);
	n = sendto(f, dp, size + 4, 0,
		   (struct sockaddr *)&peer, peer_len);
	if (n != size + 4) {
	    perror("tftp: sendto");
	    goto abort;
	}
	read_ahead(file, convert);
	for ( ; ; ) {
	    alarm(rexmtval);
	    do {
		fromlen = sizeof (from);
		n = recvfrom(f, ackbuf, sizeof (ackbuf), 0,
			     (struct sockaddr *)&from, &fromlen);
	    } while (n <= 0);
	    alarm(0);
	    if (n < 0) {
		perror("tftp: recvfrom");
		goto abort;
	    }
	    switch (((struct sockaddr *)&peer)->sa_family) {
	    case AF_INET:
		((struct sockaddr_in *)&peer)->sin_port =
		    ((struct sockaddr_in *)&from)->sin_port;
		break;
#ifdef INET6
	    case AF_INET6:
		((struct sockaddr_in6 *)&peer)->sin6_port =
		    ((struct sockaddr_in6 *)&from)->sin6_port;
		break;
#endif
	    default:
		/* unsupported */
		break;
	    }
	    if (trace)
		tpacket("received", ap, n);
	    /* should verify packet came from server */
	    ap->th_opcode = ntohs(ap->th_opcode);
	    ap->th_block = ntohs(ap->th_block);
	    if (ap->th_opcode == OACK) {
		break;
	    }
	    if (ap->th_opcode == ERROR) {
		printf("Error code %d: %s\n", ap->th_code,
		       ap->th_msg);
		goto abort;
	    }
	    if (ap->th_opcode == ACK) {
		volatile int j = 0;

		if (ap->th_block == block) {
		    break;
		}
		/* On an error, try to synchronize
		 * both sides.
		 */
		j = synchnet(f);
		if (j && trace) {
		    printf("discarded %d packets\n",
			   j);
		}
		if (ap->th_block == (block-1)) {
		    goto send_data;
		}
	    }
	}
	if (firsttrip) {
	    firsttrip = 0;
	}
	else {
	    amount += size;
	}
	block++;
    } while (size == SEGSIZE || block == 1);
  abort:
    fclose(file);
    stopclock();
    if (amount > 0)
	printstats("Sent", amount);
}

/*
 * Receive a file.
 */
void
recvfile(int fd, char *name, char *mode)
{
    register struct tftphdr *ap;
    struct tftphdr *dp;
    volatile int size = 0;
    volatile u_int16_t block = 1;
    int n;
    volatile unsigned long amount = 0;
    struct sockaddr_storage peer, from;
    volatile int firsttrip = 1;
    FILE *file;
    volatile int convert;            /* true if converting crlf -> lf */

    startclock();
    dp = w_init();
    ap = (struct tftphdr *)ackbuf;
    file = fdopen(fd, "w");
    convert = !strcmp(mode, "netascii");
    memcpy(&peer, &peeraddr, peer_len);

    mysignal(SIGALRM, timer);
    do {
	if (firsttrip) {
	    size = makerequest(RRQ, name, ap, mode);
	    firsttrip = 0;
	} else {
            ap->th_opcode = htons((u_short)ACK);
            ap->th_block = htons((u_short)(block));
            size = 4;
            block++;
        }
        timeout = 0;
        (void) sigsetjmp(timeoutbuf, 1);
send_ack:
        if ( trace )
            tpacket("sent", ap, size);


	if (sendto(f, ackbuf, size, 0, (struct sockaddr *)&peer,
		   peer_len) != size) {
            alarm(0);
	    perror("tftp: sendto");
            goto abort;
        }
        write_behind(file, convert);

        change_port(&from, &peer);

	for ( ; ; ) {

            alarm(rexmtval);

            do {
                INIT_MSG_STR( dp, PKTSIZE, from, 0 );
                n = recvmsg(f, &msgh, 0);

                if( check_src( &from ) )
                    n = -1;

            } while ( n <= 0 );

            alarm(0);

            if ( n < 0 ) {
                perror("tftp: recvfrom");
                goto abort;
            }
            if ( trace )
                tpacket("received", dp, n);
            /* should verify client address */
            dp->th_opcode = ntohs(dp->th_opcode);
            dp->th_block = ntohs(dp->th_block);

            /* If the received packet is not error packet,
               only then overwrite the peer port with the
               peer TID
            */
            if ( dp->th_opcode != ERROR )
                change_port(&peer, &from);

	    if (dp->th_opcode == OACK) {
		char *var;

		var = dp->th_data;
		if (strcasecmp(var, "ksize") == 0) {
		    TI_reqsize = atoi(var + 6);
		    if (verbose)
			printf("SIZE set to: %d\n", TI_reqsize);
		}

		ap->th_opcode = htons((u_short)ACK); /* acknowledge OACK */
		ap->th_block = htons((u_short)(block-1));
		if (sendto(f, ackbuf, 4, 0, (struct sockaddr *)&peer,
			   peer_len) != 4) {
		    alarm(0);
		    perror("tftp: sendto OACK ACK");
		    goto abort;
		}
		continue;
	    }
        if ( dp->th_opcode == ERROR ) {
            DBG_PRINT(("==>[%s:%d]: dp->th_opcode == ERROR; dp->th_code = %d: %s\n",\
                      __func__,__LINE__, (int)dp->th_code,  dp->th_msg ))
            TI_error = dp->th_code;
            /* On all errors received - report error and exit */
            goto abort;
        }
	    if (dp->th_opcode == DATA) {
            volatile int j = 0;

            TIplug.tftp_report_config(REP_SET,NULL);
		if (dp->th_block == block) {
		    break;          /* have next packet */
		}
		/* On an error, try to synchronize
		 * both sides.
		 */
		j = synchnet(f);
		if (j && trace) {
		    printf("discarded %d packets\n", j);
		}
		if (dp->th_block == (block-1)) {
		    goto send_ack;  /* resend ack */
		}
	    }
	}
	/*      size = write(fd, dp->th_data, n - 4); */
	size = writeit(file, &dp, n - 4, convert);
	if (size < 0) {
	    nak(errno + 100, &peer);
	    break;
	}
	amount += size;
    } while (size == TI_reqsize);
    ap->th_opcode = htons((u_short)ACK);    /* has seen err msg */
    ap->th_block = htons((u_short)block);
    (void) sendto(f, ackbuf, 4, 0, (struct sockaddr *)&peer, peer_len);

    write_behind(file, convert);            /* flush last buffer */
    fclose(file);
    stopclock();
    if ( amount > 0 )
        printstats("Received", amount);

    if ( TI_error == 0 )
        TIplug.tftp_report_config(REP_SET, "File received");

    return;
abort:
    fclose(file);
    stopclock();
    TIplug.tftp_report_config(REP_SET + TI_error, "Error occured");
}

int
makerequest(int request, char *name, struct tftphdr *tp, char *mode)
{
    register char *cp;

    tp->th_opcode = htons((u_short)request);
    cp = tp->th_stuff;
    strcpy(cp, name);
    cp += strlen(name);
    *cp++ = '\0';
    strcpy(cp, mode);
    cp += strlen(mode);
    *cp++ = '\0';
    strcpy(cp, "blksize");
    cp += strlen("blksize");
    *cp++ = '\0';

    if (TI_ipv6) {
        if (TI_blocksize > MAXBLKSZV6) {
            fprintf(stderr,"Bad option -b: max blocksize IPv6 %d\n", MAXBLKSZV6);
            TIplug.tftp_report_config(REP_SET + 253, "Bad option");
            exit(0);
        }
        if (TI_blocksize == 0)
            TI_blocksize = MAXBLKSZV6;

    } else {
        if (TI_blocksize > MAXBLKSZV4) {
            fprintf(stderr,"Bad option -b: max blocksize IPv4 %d\n", MAXBLKSZV4);
            TIplug.tftp_report_config(REP_SET + 253, "Bad option");
            exit(0);
        }
        if (TI_blocksize == 0)
            TI_blocksize = MAXBLKSZV4;
    }

    sprintf(cp, "%d", TI_blocksize);
    cp += 5;


    return (cp - (char *)tp);
}

struct errmsg {
    int e_code;
    const char *e_msg;
} errmsgs[] = {
    { EUNDEF,	"Undefined error code" },
    { ENOTFOUND,	"File not found" },
    { EACCESS,	"Access violation" },
    { ENOSPACE,	"Disk full or allocation exceeded" },
    { EBADOP,	"Illegal TFTP operation" },
    { EBADID,	"Unknown transfer ID" },
    { EEXISTS,	"File already exists" },
    { ENOUSER,	"No such user" },
    { -1,		0 }
};

/*
 * Send a nak packet (error message).
 * Error code passed in is one of the
 * standard TFTP codes, or a UNIX errno
 * offset by 100.
 */
void
nak(int error, struct sockaddr_storage *peer)
{
    register struct errmsg *pe;
    register struct tftphdr *tp;
    int length;

    tp = (struct tftphdr *)ackbuf;
    tp->th_opcode = htons((u_short)ERROR);
    tp->th_code = htons((u_short)error);
    for (pe = errmsgs; pe->e_code >= 0; pe++)
	if (pe->e_code == error)
	    break;
    if (pe->e_code < 0) {
	pe->e_msg = strerror(error - 100);
	tp->th_code = EUNDEF;
    }
    strcpy(tp->th_msg, pe->e_msg);
    length = strlen(pe->e_msg) + 4;
    if (trace)
	tpacket("sent", tp, length);
    if (sendto(f, ackbuf, length, 0, (struct sockaddr *)peer,
	       peer_len) != length)
	perror("nak");
}

static
void
tpacket(const char *s, struct tftphdr *tp, int n)
{
    static const char *opcodes[] =
	{ "#0", "RRQ", "WRQ", "DATA", "ACK", "ERROR", "OACK" };
    register char *cp, *file;
    u_short op = ntohs(tp->th_opcode);

    if (op < RRQ || op > OACK)
	printf("%s opcode=%x ", s, op);
    else
	printf("%s %s ", s, opcodes[op]);
    switch (op) {

        case RRQ:
        case WRQ:
            n -= 2;
            file = cp = tp->th_stuff;
            cp = cp + strlen(cp);
            printf("<file=%s, mode=%s>\n", file, cp + 1);
            break;

        case DATA:
            printf("<block=%u, %d bytes>\n", ntohs(tp->th_block), n - 4);
            break;

        case ACK:
            printf("<block=%u>\n", ntohs(tp->th_block));
            break;

        case ERROR:
            printf("<code=%d, msg=%s>\n", ntohs(tp->th_code), tp->th_msg);
            break;

        case OACK:
            file = cp = tp->th_stuff;
            cp = cp + strlen(cp);
            printf("<option=%s, value=%s>\n", file, cp + 1);
            break;

    }
}

struct timeval tstart;
struct timeval tstop;
struct timezone zone;

int get_sysUpTime(unsigned int *sysUpTimeVal)
{
    FILE *fp;
    char line[MAX_LINE_SIZE];
    char *ret_val;
    unsigned int upTime = 0;

    /* This file contains two numbers:
     * the uptime of the system (seconds), and the amount of time spent in idle process (seconds).
     * We care only for the first one */
    fp = fopen( UPTIME_FILE_PATH, "r");
    if ( fp == NULL )
    {
        return -1;
    }

    ret_val = fgets(line,MAX_LINE_SIZE,fp);
    fclose(fp);

    if ( ret_val == NULL )
    {
        return -1;
    }

    /* Extracting the first token (number of up-time in seconds). */
    ret_val = strtok (line," .");

    /* we need only the number of seconds */
    upTime += atoi(ret_val);

    *sysUpTimeVal = upTime;

    return 0;
}

void
startclock(void) {
    get_sysUpTime(&tstart.tv_sec);
    srand(tstart.tv_usec);
}

void
stopclock(void) {
    get_sysUpTime(&tstop.tv_sec);
}

void
printstats(const char *direction, unsigned long amount)
{
    double delta;
    /* compute delta in 1/10's second units */
    delta = ((tstop.tv_sec*10.)+(tstop.tv_usec/100000)) -
	((tstart.tv_sec*10.)+(tstart.tv_usec/100000));
    delta = delta/10.;      /* back to seconds */
    printf("%s %ld bytes in %.1f seconds", direction, amount, delta);
    if (verbose)
	printf(" [%.0f bits/sec]", (amount*8.)/delta);
    putchar('\n');
}

